"""Fetch plugin info such as aliases and __doc__."""
# pylint: disable=invalid-name
from typing import List, Union
from types import ModuleType

from pathlib import Path
import re
import ast
import logzero
from logzero import logger

import nonebot

logzero.loglevel(10)


def extrac_info(file: Union[str, Path]) -> str:
    """Extract command/aliases info for a given a file."""
    if not Path(file).is_file():
        logger.warning(" file [%s] does not exist or is not a file.", file)
        raise Exception(f" file [{file}] does not exist or is not a file.")

    try:
        text = Path(file).read_text(encoding="utf8")
    except Exception as e:
        logger.error(e)
        raise

    # guess = on_command(...)
    _ = re.findall(r"^[^#]*\s*=\s+(?:nonebot\.)?on_command\([\s\S]*?\)", text, re.M)
    if _:
        str_ = _[0]
    else:
        str_ = ""

    res = ""
    # command
    _ = re.findall(r"""^[\s\S]*?["']([^"']+)""", str_)
    if _:
        res += f"\ncommand: {_[0]}\n"

    # aliases
    _ = re.findall(r"aliases\s*=\s*\{([\s\S]*)\}", str_, re.M)
    if _:
        aliases = ast.literal_eval(_[0])
        aliases = ", ".join(aliases)
        if aliases:
            res += f"\taliases: {aliases}\n"

    return res


def fetch_aliases(mod: ModuleType, attach_doc: bool = False) -> str:
    """Fetch command and aliases if present.

    import nonebot_plugin_guess
    _fetch_aliases(nonebot_plugin_guess)
    """
    if Path(mod.__file__).name in ["__init__.py"]:
        # is a package, search the folder for command/aliases info
        p_dir = Path(mod.__file__).parent
        info = ""
        for file in p_dir.glob("*.py"):
            try:
                info = extrac_info(file)
            except Exception as e:
                logger.error("file: %s, extrac_info(file) exc: %s", file, e)
                info = ""
            if info.strip():
                break
    else:
        try:
            info = extrac_info(mod.__file__)
        except Exception as e:
            logger.error(
                "mod.__file__: %s, extrac_info(mod.__file__) exc: %s", mod.__file__, e
            )
            info = ""

    info = info.strip()
    if attach_doc:
        if mod.__doc__ is None or not mod.__doc__.strip():
            info += f'\n(无文档。可在{mod.__file__}开始处以"""..."""形式添加文档'
        else:
            info += f"\n__doc__:\n{mod.__doc__}"

    return info


def fetch_plugin_info(
    # mod_list: Optional[Union[ModuleType, List[ModuleType]]] = None,
    mod_list: List[ModuleType] = None,
    details: bool = False,
) -> str:
    """Fetch plugin info such as aliases and __doc__.

    Args:
        mod_list: target modules
        details: whether to attach moduel.__doc__

    Returns:
        command and it's aliases, and module.__doc__ if details is True
    """
    if mod_list is None:
        mod_list = [elm.module for elm in nonebot.get_loaded_plugins()]

    # if isinstance(mod_list, ModuleType): mod_list = list(mod_list)

    res = ""
    for module in mod_list:
        res += "\n" + fetch_aliases(module, attach_doc=details)

    return res



"""Autogenerate response to commands /help !help help.
/help details: also print __doc__

For more info on usage:
/help -h
"""
# pylint: disable=invalid-name, too-many-statements, no-name-in-module, too-many-locals
# import asyncio

from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter

import re
from time import time
import logzero
from logzero import logger

import nonebot
from nonebot.typing import T_State
from nonebot.params import State

# from nonebot.adapters.cqhttp import Bot, Event
from nonebot.adapters.onebot.v11 import Bot, Event

from .parse_cmd import parse_cmd
from .fetch_plugin_info import fetch_plugin_info

# for ratelimit, will not respond twice within 3 seconds
_vars = dict(last_sent=0.0, interval=3)
patt = re.compile(r"^[/!#]?\s*(?:help|menu|帮助|菜单|caidan|info)|[/#]i", re.I)

# logzero.loglevel(20) to disable noisy debug messages
logzero.loglevel(10)

# on_message = nonebot.on_message(priority=1)
nonebot_plugin_autohelp = nonebot.on_message(priority=1, block=False)

config = nonebot.Config()

# nonebot.load_from_toml("pyproject.toml")
# nonebot.load_plugin("nonebot_plugin_guess")

# wait for other plugins to load
# async def wait_for_30(): await asyncio.sleep(30)

logger.info("Loaded plugins: %s", [elm.name for elm in nonebot.get_loaded_plugins()])

# module path (__init__.py or plugin_name.py path)
# [elm.module.__file__ for elm in nonebot.get_loaded_plugins()]

# module doc: info at the begenning of the file
# [elm.module.__doc__ for elm in nonebot.get_loaded_plugins()]

# func doc: elm.module.handle.__doc__


@nonebot_plugin_autohelp.handle()
# async def handle(bot: Bot, event: Event, state: dict):
async def handle(bot: Bot, event: Event, state: T_State = State()):
    """Handle messages."""
    logger.debug(" nonebot_plugin_autohelp entry ")
    logger.debug("state: %s", state)

    _ = time() - _vars.get("last_sent", 0)
    logger.debug("check time interval: %.1f", _)
    if _ < _vars["interval"]:
        logger.debug("Too soon... return ...")
        return

    # messages not startswith [/!#]?\s*help (case-insensitive
    msg = event.get_plaintext()
    logger.debug("msg: %s", msg)
    # if not re.findall(, msg, re.I):
    if not patt.findall(msg):
        logger.debug("patt.findall(msg) False, return...")
        return

    parser = ArgumentParser(
        prog="help",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "-d", "--details", action="store_true", help="show __doc__ for each plugin"
    )
    parser.add_argument("params", nargs="*", help="list of parameters of type str")

    command = str(event.get_message()).strip()
    logger.debug("command (str(event.message).strip()): %s", command)

    args, stdout, stderr = parse_cmd(command, parser)
    logger.debug("args: %s", args)

    if stdout or stderr:
        await bot.send(message="\n---\n".join([stdout, stderr]), event=event)
        return

    keys = [
        "nickname",
        "command_start",
        "command_sep",
    ]
    info = "\n".join(
        f"{key}: {', '.join(val)}" for key, val in nonebot.Config() if key in keys
    )
    logger.debug("Respond... \n%s", info)

    logger.debug("args: %s", args)

    # args.params contains "details" or "detail" or "详细"
    det = any(map(lambda x: x in args.params, ["details", "detail", "详细"]))

    try:
        args_details = args.details
    except AttributeError:
        args_details = False

    if args_details or det:
        try:
            plugin_info = fetch_plugin_info(details=True)
        except Exception as e:
            logger.error("fetch_plugin_info() exc: %s", e)
            plugin_info = str(e)
        try:
            await bot.send(message=f"{info}\n{plugin_info}", event=event)

            # reset timer if sent successfully
            _vars["last_sent"] = time()
        except Exception as e:
            logger.error(e)
    else:
        try:
            plugin_info = fetch_plugin_info()
        except Exception as e:
            logger.error("fetch_plugin_info() exc: %s", e)
            plugin_info = str(e)
        try:
            await bot.send(
                message=f"{info}\n{plugin_info}\n(help -d will display detailed docs for all plugins loaded before nonebot_plugin_autohelp)",
                event=event,
            )

            # reset timer if sent successfully
            _vars["last_sent"] = time()
        except Exception as e:
            logger.error(e)



"""Parse command according to a given argparse.ArgumentParser."""
# pylint: disable=invalid-name

from typing import Tuple

from argparse import ArgumentParser, Namespace, ArgumentDefaultsHelpFormatter
import shlex
from contextlib import redirect_stderr, redirect_stdout
from io import StringIO

from logzero import logger


def parse_cmd(
    command: str, parser: ArgumentParser = None
) -> Tuple[Namespace, str, str]:
    """Parse command a given argparse.ArgumentParser.

    Args
        command: str to parse
        parser: predefined template
    Returns
        args, stdout, stderr
    """
    if parser is None:
        parser = ArgumentParser(
            prog="dummy",
            description='dummy parser for cases when no parser is provided',
            # exit_on_error=False,  # exit_on_error avail in 3.9
            formatter_class=ArgumentDefaultsHelpFormatter,
        )
        parser.add_argument("params", nargs="*", help="list of parameters of type str")

    # capture stderr
    stderr = StringIO()
    stdout = StringIO()
    # args: Optional[Namespace] = None
    args = parser.parse_args([])
    with redirect_stderr(stderr), redirect_stdout(stdout):
        # catch SystemExit
        try:
            # ignore the first entry that's prg, just need args
            args = parser.parse_args(shlex.split(command)[1:])
        except SystemExit:
            logger.error("SystemExit caught.")
        except BaseException as e:
            logger.error(e)
    if stderr.getvalue():
        logger.error(stderr.getvalue())

    return args, stdout.getvalue(), stderr.getvalue()



"""Autogenerate response to commands /help !help help ²Ëµ¥ caidan /i.

For more info on usage:
/help -h
"""
__version__ = "0.1.5"
from .nonebot_plugin_autohelp import nonebot_plugin_autohelp

__all__ = ("nonebot_plugin_autohelp",)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nonebot_plugin_autohelp']

package_data = \
{'': ['*']}

install_requires = \
['logzero>=1.7.0,<2.0.0',
 'nonebot-adapter-onebot>=2.0.0-beta.1,<3.0.0',
 'nonebot2>=2.0.0-alpha.16,<3.0.0']

setup_kwargs = {
    'name': 'nonebot-plugin-autohelp',
    'version': '0.1.5',
    'description': 'nonebot2 plugin 自动生成载入插件的帮助信息，响应 /help !help help 菜单 caidan /i',
    'long_description': '# nonebot-plugin-autohelp\n[![nonebot2beta](https://img.shields.io/static/v1?label=nonebot&message=v2b2&color=green)](https://v2.nonebot.dev/)[![onebot](https://img.shields.io/static/v1?label=driver&message=onebot&color=green)](https://adapter-onebot.netlify.app/)[![python](https://img.shields.io/static/v1?label=python+&message=3.7%2B&color=blue)](https://img.shields.io/static/v1?label=python+&message=3.7%2B&color=blue)[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)[![PyPI version](https://badge.fury.io/py/nonebot_plugin_autohelp.svg)](https://badge.fury.io/py/nonebot_plugin_autohelp)\n\nShow a summary of commands, aliases and usages for plugins loaded\n\n## Dependent adapter\n`onebotv11`\n\n## Install it\n\n```shell\npip install nonebot-plugin-autohelp\n# or poetry add nonebot-plugin-autohelp\n# pip install git+htts://github.com/ffreemt/nonebot-plugin-autohelp\n# poetry add git+htts://github.com/ffreemt/nonebot-plugin-autohelp\n\n# To upgrade\npip install nonebot-plugin-autohelp -U\n# or poetry add nonebot-plugin-autohelp@latest\n```\n\n## Use it\n```python\n# bot.py\nimport nonebot\nfrom nonebot.adapters.onebot.v11 import Adapter\n...\nnonebot.init()\n\ndriver = nonebot.get_driver()\ndriver.register_adapter(Adapter)\n\nnonebot.load_from_toml("pyproject.toml")\nnonebot.load_builtin_plugin("echo")\nnonebot.load_plugin("nonebot_plugin_guess")\n\nnonebot.load_plugin("nonebot_plugin_autohelp")\n\n# plugin loaded after autohelp will not be taken care of by autohelp\nnonebot.load_plugin("nonebot_plugin_fancy")\n\n```\n\nSample session in a qq group\n```bash\nmu (μ)(41947782)  11:23:34 AM\nhelp\nmubot(2129462094)  11:23:36 AM\nnickname:\ncommand_start: /\ncommand_sep: .\n\ncommand: say\ncommand: mecho\n\taliases: ping, ryt, 在不, p\n\ncommand: news\n\taliases: xinwen, 新闻, 无聊\n\ncommand: debug test: %s\n\taliases: 爬, fetch, crawl\n\ncommand: guess\n\taliases: cai, 猜猜看, 猜\n\n(help -d will display detailed docs for all plugins loaded before nonebot_plugin_autohelp)\n\nmu (μ)(41947782)  11:53:25 AM\nhelp -h\nmubot(2129462094)  11:53:27 AM\nusage: help [-h] [-d] [params [params ...]]\n\npositional arguments:\n  params         list of parameters of type str (default: None)\n\noptional arguments:\n  -h, --help     show this help message and exit\n  -d, --details  show __doc__ for each plugin (default: False)\n---\n\nmu (μ)(41947782)  11:23:34 AM\nhelp --details  # or help details\n...(attach __doc___ for each plugin, ommitted)\n```\n\n',
    'author': 'ffreemt',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ffreemt/nonebot-plugin-autohelp',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.3,<4.0.0',
}


setup(**setup_kwargs)
